@@ -43,21 +43,6 @@
 @end
 
 
-/* 
-    Structure for rez switch gamma fades
-    We can hide the monitor flicker by setting the gamma tables to 0
-*/
-#define QZ_GAMMA_TABLE_SIZE 256
-
-typedef struct {
-
-    CGGammaValue red[QZ_GAMMA_TABLE_SIZE];
-    CGGammaValue green[QZ_GAMMA_TABLE_SIZE];
-    CGGammaValue blue[QZ_GAMMA_TABLE_SIZE];
-
-} SDL_QuartzGammaTable;
-
-
 /* Bootstrap functions */
 static int              QZ_Available ();
 static SDL_VideoDevice* QZ_CreateDevice (int device_index);
@@ -184,12 +169,18 @@ static void QZ_DeleteDevice (SDL_VideoDevice *device) {
 
 static int QZ_VideoInit (_THIS, SDL_PixelFormat *video_format) {
 
+    NSRect r = NSMakeRect(0.0, 0.0, 0.0, 0.0);
+    const char *env = NULL;
+
     /* Initialize the video settings; this data persists between mode switches */
     display_id = kCGDirectMainDisplay;
     save_mode  = CGDisplayCurrentMode    (display_id);
     mode_list  = CGDisplayAvailableModes (display_id);
     palette    = CGPaletteCreateDefaultColorPalette ();
 
+    env = SDL_getenv("SDL_VIDEO_ALLOW_SCREENSAVER");
+    allow_screensaver = ( env && SDL_atoi(env) ) ? YES : NO;
+
     /* Gather some information that is useful to know about the display */
     CFNumberGetValue (CFDictionaryGetValue (save_mode, kCGDisplayBitsPerPixel),
                       kCFNumberSInt32Type, &device_bpp);
@@ -212,6 +203,7 @@ static int QZ_VideoInit (_THIS, SDL_PixelFormat *video_format) {
     cursor_should_be_visible    = YES;
     cursor_visible              = YES;
     current_mods = 0;
+    field_edit =  [[NSTextView alloc] initWithFrame:r];
     
     if ( Gestalt(gestaltSystemVersion, &system_version) != noErr )
         system_version = 0;
@@ -375,6 +367,13 @@ static void QZ_UnsetVideoMode (_THIS, BOOL to_desktop) {
             SDL_free (sw_buffers[0]);
         }
         
+        /* If we still have a valid window, close it. */
+        if ( qz_window ) {
+            [ qz_window close ];
+            [ qz_window release ];
+            qz_window = nil;
+            window_view = nil;
+        }
         /* 
             Release the OpenGL context
             Do this first to avoid trash on the display before fade
@@ -385,10 +384,10 @@ static void QZ_UnsetVideoMode (_THIS, BOOL to_desktop) {
             CGLSetFullScreen (NULL);
         }
         if (to_desktop) {
+            ShowMenuBar ();
             /* Restore original screen resolution/bpp */
             CGDisplaySwitchToMode (display_id, save_mode);
             CGReleaseAllDisplays ();
-            ShowMenuBar ();
             /* 
                 Reset the main screen's rectangle
                 See comment in QZ_SetVideoFullscreen for why we do this
@@ -419,6 +418,8 @@ static SDL_Surface* QZ_SetVideoFullScreen (_THIS, SDL_Surface *current, int widt
     boolean_t exact_match = 0;
     NSRect screen_rect;
     CGError error;
+    NSRect contentRect;
+    BOOL isCustom = NO;
     CGDisplayFadeReservationToken fade_token = kCGDisplayFadeReservationInvalidToken;
     
     /* Fade to black to hide resolution-switching flicker (and garbage
@@ -511,6 +512,59 @@ static SDL_Surface* QZ_SetVideoFullScreen (_THIS, SDL_Surface *current, int widt
     if ( CGDisplayCanSetPalette (display_id) )
         current->flags |= SDL_HWPALETTE;
 
+    /* The code below checks for any valid custom windows and views.  If none are
+       available, then we create new ones.  Window/View code was added in FULLSCREEN
+       so that special events like the changing of the cursor image would be handled
+       ( only the front-most and active application can change the cursor appearance
+       and with no valid window/view in FULLSCREEN, SDL wouldn't update its cursor. )
+    */
+	/* Check for user-specified window and view */
+    {
+        char *windowPtrString = getenv ("SDL_NSWindowPointer");
+        char *viewPtrString = getenv ("SDL_NSQuickDrawViewPointer");
+    
+        contentRect = NSMakeRect (0, 0, width, height);
+	
+        if (windowPtrString && viewPtrString) {
+            /* Release any previous window */
+            if ( qz_window ) {
+                [ qz_window release ];
+                qz_window = nil;
+            }
+            
+            qz_window = (NSWindow*)atoi(windowPtrString);
+            window_view = (NSQuickDrawView*)atoi(viewPtrString);
+            isCustom = YES;
+            /* 
+                Retain reference to window because we
+                might release it in QZ_UnsetVideoMode
+            */
+            [ qz_window retain ];
+        }
+    }
+    /* Check if we should recreate the window */
+    if (qz_window == nil) {
+        /* Manually create a window, avoids having a nib file resource */
+        qz_window = [ [ SDL_QuartzWindow alloc ] 
+            initWithContentRect:contentRect
+                styleMask:nil 
+                    backing:NSBackingStoreBuffered
+                        defer:NO ];
+
+        if (qz_window != nil) {
+            [ qz_window setAcceptsMouseMovedEvents:YES ];
+            [ qz_window setViewsNeedDisplay:NO ];
+        }
+    }
+    /* We already have a window, just change its size */
+    else {
+        if (!isCustom) {
+            [ qz_window setContentSize:contentRect.size ];
+            current->flags |= (SDL_NOFRAME|SDL_RESIZABLE) & mode_flags;
+            [ window_view setFrameSize:contentRect.size ];
+        }
+    }
+
     /* Setup OpenGL for a fullscreen context */
     if (flags & SDL_OPENGL) {
 
@@ -521,6 +575,12 @@ static SDL_Surface* QZ_SetVideoFullScreen (_THIS, SDL_Surface *current, int widt
             goto ERR_NO_GL;
         }
 
+        /* Initialize the NSView and add it to our window.  The presence of a valid window and
+           view allow the cursor to be changed whilst in fullscreen.*/
+        window_view = [ [ NSView alloc ] initWithFrame:contentRect ];
+        [ [ qz_window contentView ] addSubview:window_view ];	
+        [ window_view release ];
+
         ctx = [ gl_context cglContext ];
         err = CGLSetFullScreen (ctx);
 
@@ -696,6 +756,7 @@ static SDL_Surface* QZ_SetVideoWindowed (_THIS, SDL_Surface *current, int width,
         }
         [ qz_window setDelegate:
             [ [ [ SDL_QuartzWindowDelegate alloc ] init ] autorelease ] ];
+        [ qz_window setContentView: [ [ [ SDL_QuartzView alloc ] init ] autorelease ] ];
     }
     /* We already have a window, just change its size */
     else {
@@ -1466,6 +1527,11 @@ static void QZ_VideoQuit (_THIS) {
         opengl_library = NULL;
     }
     this->gl_config.driver_loaded = 0;
+
+    if (field_edit) {
+        [field_edit release];
+        field_edit = NULL;
+    }
 }
 
 #if 0 /* Not used (apparently, it's really slow) */
